/*
    WinCdArb - Windows CdRom Device Arbiter service

    Copyright (C) 2007-2019 GuinpinSoft inc <wincdarb@makemkv.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/
#include "winntapi.h"
#include <wincdarb/cdarbdefs.h>
#include "cdarbsrv.h"

#include <ver_num.h>

#if defined(_M_IX86)
#define WINCDARB_ARCH "x86"
#elif defined(_M_AMD64)
#define WINCDARB_ARCH "x64"
#elif defined(_M_ARM)
#define WINCDARB_ARCH "arm"
#else
#define WINCDARB_ARCH "unk"
#endif

static const CHAR exePath[] = "cdarbsvc";
static const CHAR exeName[] = "cdarbsvc_" WINCDARB_VERSION_NUMBER "_" WINCDARB_ARCH ".exe";

static const size_t exePathLength = sizeof(exePath)-1;
static const size_t exeNameLength = sizeof(exeName)-1;

typedef SC_HANDLE   (*AccessServiceOrSCMFunction)(SC_HANDLE scm,ULONG_PTR Param);
typedef HRESULT     (*ManageServiceOrSCMFunction)(SC_HANDLE svc,ULONG_PTR Param);
typedef HRESULT     (*ManageServiceFunction)(SC_HANDLE svc);

static HRESULT ManageServiceOrSCM(SC_HANDLE scm,ULONG_PTR AccessParam,ULONG_PTR ManageParam,AccessServiceOrSCMFunction AccessProc,ManageServiceOrSCMFunction ManageProc)
{
    HRESULT         hr;
    SC_HANDLE       svc;

    svc = (*AccessProc)(scm,AccessParam);

    if (NULL==svc)
    {
        return HRESULT_FROM_WIN32(GetLastError());
    }

    hr = (*ManageProc)(svc,ManageParam);

    CloseServiceHandle(svc);

    return hr;
}

static inline HRESULT ManageCdArbService(SC_HANDLE scm,DWORD dwDesiredAccess,ManageServiceFunction ManageProc)
{
    return ManageServiceOrSCM(scm,dwDesiredAccess,(ULONG_PTR)(PVOID)ManageProc,
        [](SC_HANDLE scm,ULONG_PTR ulpDesiredAccess)
    {
        return OpenServiceW(scm,
            WINCDARB_SERVICE_NAME,
            (DWORD)ulpDesiredAccess);
    },
        [](SC_HANDLE svc,ULONG_PTR ulpManageProc)
    {
        return (*(ManageServiceFunction)(PVOID)ulpManageProc)(svc);
    });
}

static inline HRESULT ManageSCM(DWORD dwDesiredAccess,ULONG_PTR Param,ManageServiceOrSCMFunction ManageProc)
{
    return ManageServiceOrSCM(NULL,dwDesiredAccess,Param,
        [](SC_HANDLE /*scm*/,ULONG_PTR ulpDesiredAccess)
    {
        return OpenSCManager(NULL,NULL,
            SC_MANAGER_CONNECT | ((DWORD)ulpDesiredAccess) );
    },
    ManageProc);
}

static HRESULT StopAndWaitService(SC_HANDLE svc)
{
    HRESULT         hr;
    DWORD           sz;
    bool            stopped;
    union _status
    {
        SERVICE_STATUS_PROCESS  ssp;
        SERVICE_STATUS          ss;
    } status;

    stopped = false;
    for (unsigned int i=0; i<200; i++)
    {
        if (FALSE==QueryServiceStatusEx(svc,SC_STATUS_PROCESS_INFO,(PBYTE)&status.ssp,sizeof(status.ssp),&sz))
        {
            hr = HRESULT_FROM_WIN32(GetLastError());
            CloseServiceHandle(svc);
            return hr;
        }
        if (SERVICE_STOPPED==status.ssp.dwCurrentState)
        {
            stopped = true;
            break;
        }

        if (FALSE!=ControlService(svc,SERVICE_CONTROL_STOP,&status.ss))
        {
            if (SERVICE_STOPPED==status.ss.dwCurrentState)
            {
                stopped = true;
                break;
            }
        }

        Sleep(100);
    }

    if (false==stopped)
    {
        return E_FAIL;
    }

    return S_OK;
}

static inline HRESULT DeleteCdArbService(SC_HANDLE scm)
{
    return ManageCdArbService(scm,
        DELETE,[](SC_HANDLE svc)
    {
        if (FALSE==DeleteService(svc))
        {
            return HRESULT_FROM_WIN32(GetLastError());
        }
        return S_OK;
    });
}

static inline HRESULT StopAndDeleteCdArbService(SC_HANDLE scm)
{
    return ManageCdArbService(scm,
        SERVICE_STOP|SERVICE_QUERY_STATUS|DELETE,
        [](SC_HANDLE svc)
    {
        HRESULT hr;

        hr = StopAndWaitService(svc);
        if (S_OK!=hr) return hr;

        if (FALSE==DeleteService(svc))
        {
            return HRESULT_FROM_WIN32(GetLastError());
        }
        return S_OK;
    });
}

static inline HRESULT StartCdArbService(SC_HANDLE scm)
{
    return ManageCdArbService(scm,
        SERVICE_START,
        [](SC_HANDLE svc)
    {
        if (FALSE==StartServiceW(svc,0,NULL))
        {
            return HRESULT_FROM_WIN32(GetLastError());
        }
        return S_OK;
    });
}

static inline HRESULT StopCdArbService(SC_HANDLE scm)
{
    return ManageCdArbService(scm,
        SERVICE_STOP|SERVICE_QUERY_STATUS,
        StopAndWaitService);
}

static inline HRESULT InstallCdArbService(SC_HANDLE scm,PCWSTR exePath)
{
    return ManageServiceOrSCM(scm,
        (ULONG_PTR)(PVOID)exePath,0,
        [](SC_HANDLE scm,ULONG_PTR ulExePath)
    {
        return CreateServiceW(scm,
            WINCDARB_SERVICE_NAME,
            WinCdArbServiceDisplayName,
            SERVICE_ALL_ACCESS,
            SERVICE_WIN32_OWN_PROCESS,
            SERVICE_AUTO_START,
            SERVICE_ERROR_NORMAL,
            (PCWSTR)(PVOID)ulExePath,
            NULL,NULL,NULL,NULL,NULL);
    },
        [](SC_HANDLE svc,ULONG_PTR)
    {
        SERVICE_DESCRIPTIONW    sdesc;
        sdesc.lpDescription = (PWSTR)WinCdArbServiceDescription;
        ChangeServiceConfig2W(svc,SERVICE_CONFIG_DESCRIPTION,&sdesc);
        return S_OK;
    });
}

static HRESULT GetCdArbServiceExePath(PWSTR Buffer,DWORD BufferLength,bool PathOnly)
{
    DWORD sz;

    sz = GetEnvironmentVariableW(L"CommonProgramFiles",Buffer,BufferLength);
    if (0==sz)
    {
        return HRESULT_FROM_WIN32(GetLastError());
    }
    if ((sz>=BufferLength) || ((BufferLength-sz)<(4+exePathLength+exeNameLength)))
    {
        return E_NOT_SUFFICIENT_BUFFER;
    }

    Buffer[sz++] = '\\';
    for (size_t i=0; i<exePathLength; i++)
    {
        Buffer[sz++] = exePath[i];
    }

    if (false==PathOnly)
    {
        Buffer[sz++] = '\\';
        for (size_t i=0; i<exeNameLength; i++)
        {
            Buffer[sz++] = exeName[i];
        }
    }

    Buffer[sz] = 0;
    return S_OK;
}

static HRESULT LoadExeFile(PVOID* pData,DWORD* pSize,PCWSTR ExePath)
{
    HANDLE          h_f,h_s;
    LARGE_INTEGER   fsize;
    HRESULT         hr;
    PVOID           data;

    h_f = CreateFileW(ExePath,GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,FILE_FLAG_NO_BUFFERING,NULL);
    if (INVALID_HANDLE_VALUE==h_f)
    {
        return HRESULT_FROM_WIN32(GetLastError());
    }

    if (FALSE==GetFileSizeEx(h_f,&fsize))
    {
        hr = HRESULT_FROM_WIN32(GetLastError());
        CloseHandle(h_f);
        return hr;
    }

    if (fsize.u.HighPart!=0)
    {
        return HRESULT_FROM_WIN32(ERROR_FILE_TOO_LARGE);
    }

    h_s = CreateFileMappingW(h_f,NULL,PAGE_READONLY,0,fsize.u.LowPart,NULL);
    hr = HRESULT_FROM_WIN32(GetLastError());

    CloseHandle(h_f);

    if (NULL==h_s)
    {
        return hr;
    }

    data = MapViewOfFile(h_s,FILE_MAP_READ,0,0,fsize.u.LowPart);
    hr = HRESULT_FROM_WIN32(GetLastError());

    CloseHandle(h_s);

    if (NULL==data)
    {
        return hr;
    }

    *pData = data;
    *pSize = fsize.u.LowPart;

    return S_OK;
}

static HRESULT CompareExeFile(LPCVOID ExeData,DWORD ExeSize,HANDLE ExeHandle)
{
    LARGE_INTEGER   fsize;
    HRESULT         hr;
    DWORD           offset;
    UCHAR           buffer[1024];

    if (FALSE==GetFileSizeEx(ExeHandle,&fsize))
    {
        return HRESULT_FROM_WIN32(GetLastError());
    }

    if ((fsize.u.HighPart!=0) || (fsize.u.LowPart!=ExeSize))
    {
        return E_FAIL;
    }

    hr = S_OK;
    offset = 0;
    while (offset < ExeSize)
    {
        DWORD len,rd;

        len = ExeSize - offset;
        if (len>sizeof(buffer))
        {
            len = sizeof(buffer);
        }
        if (FALSE==ReadFile(ExeHandle,buffer,len,&rd,NULL))
        {
            return HRESULT_FROM_WIN32(GetLastError());
        }
        if ( (len!=rd) || (0!=memcmp(buffer,((LPCBYTE)ExeData)+offset,len)) )
        {
            return E_FAIL;
        }

        offset += len;
    }

    return S_OK;
}

static HRESULT WriteExeFile(LPCVOID ExeData,DWORD ExeSize,HANDLE ExeHandle)
{
    DWORD offset;

    C_ASSERT(INVALID_SET_FILE_POINTER!=0);

    if (INVALID_SET_FILE_POINTER==SetFilePointer(ExeHandle,0,NULL,FILE_BEGIN))
    {
        return HRESULT_FROM_WIN32(GetLastError());
    }

    static const unsigned int MaxWriteSize=16384;

    offset = 0;
    while (offset < ExeSize)
    {
        DWORD len,wr;

        len = ExeSize - offset;
        if (len>MaxWriteSize)
        {
            len = MaxWriteSize;
        }
        if (FALSE==WriteFile(ExeHandle,((LPCBYTE)ExeData)+offset,len,&wr,NULL))
        {
            return HRESULT_FROM_WIN32(GetLastError());
        }
        if (len!=wr)
        {
            return E_FAIL;
        }

        offset += len;
    }

    if (FALSE==SetEndOfFile(ExeHandle))
    {
        return HRESULT_FROM_WIN32(GetLastError());
    }
    FlushFileBuffers(ExeHandle);

    return S_OK;
}


static HRESULT CopyExeFile(LPCVOID ExeData,DWORD ExeSize,PCWSTR ExePath)
{
    HANDLE  h;
    HRESULT hr;

    h = CreateFileW(ExePath,GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,0,NULL);
    if (INVALID_HANDLE_VALUE!=h)
    {
        // file exist
        if (S_OK==CompareExeFile(ExeData,ExeSize,h))
        {
            // file is the same
            CloseHandle(h);
            return S_OK;
        }

        // file exist, and is invalid. should not really happen
        CloseHandle(h);
    }

    h = CreateFileW(ExePath,GENERIC_READ|GENERIC_WRITE,FILE_SHARE_READ,NULL,CREATE_ALWAYS,0,NULL);
    if (INVALID_HANDLE_VALUE==h)
    {
        return HRESULT_FROM_WIN32(GetLastError());
    }

    hr = WriteExeFile(ExeData,ExeSize,h);
    CloseHandle(h);

    return hr;
}

static HRESULT GetThisExecutablePath(PWSTR Buffer,DWORD BufferLength)
{
    DWORD   sz;

    sz=GetModuleFileNameW(NULL,Buffer,BufferLength);
    if ((0==sz) || (sz>=BufferLength))
    {
        return HRESULT_FROM_WIN32(ERROR_INSUFFICIENT_BUFFER);
    }

    return S_OK;
}

static HRESULT MakeTinyExecutablePath(PWSTR Buffer)
{
    size_t  len;
    PWCHAR  p;
    DWORD   attr;

    len = wcslen(Buffer);
    if (len<12) return E_INVALIDARG;

    p = NULL;
    if (0==memcmp(Buffer + (len-9),L"_inst.exe",9*sizeof(WCHAR)))
    {
        p = Buffer + (len-9);
    }
    if (0==memcmp(Buffer + (len-11),L"_inst64.exe",11*sizeof(WCHAR)))
    {
        p = Buffer + (len-11);
    }
    if (NULL==p) return E_INVALIDARG;

    memcpy(p+1,L"tiny",4*sizeof(WCHAR));

    attr = GetFileAttributesW(Buffer);

    if ((INVALID_FILE_ATTRIBUTES==attr) ||
        (0!=(FILE_ATTRIBUTE_DIRECTORY&attr)))
    {
        return HRESULT_FROM_WIN32(ERROR_FILE_NOT_FOUND);
    }

    return S_OK;
}


static HRESULT CopyCdArbExecutable(PWSTR Buffer,DWORD BufferLength,bool UseTinyExe)
{
    HRESULT hr;
    PVOID   exeData;
    DWORD   exeSize;

    hr = GetCdArbServiceExePath(Buffer,BufferLength,true);
    if (S_OK!=hr) return hr;

    // create directory
    if (0==CreateDirectoryW(Buffer,NULL))
    {
        hr = HRESULT_FROM_WIN32(GetLastError());
    } else {
        hr = S_OK;
    }
    if (HRESULT_FROM_WIN32(ERROR_ALREADY_EXISTS)==hr)
    {
        hr=S_OK;
    }
    if (S_OK!=hr) return hr;

    hr = GetThisExecutablePath(Buffer,BufferLength);
    if (S_OK!=hr) return hr;

    if (UseTinyExe)
    {
        hr = MakeTinyExecutablePath(Buffer);
        if (S_OK!=hr) return hr;
    }

    hr = LoadExeFile(&exeData,&exeSize,Buffer);
    if (S_OK!=hr) return hr;

    hr = GetCdArbServiceExePath(Buffer,BufferLength,false);
    if (S_OK!=hr) return hr;

    hr = CopyExeFile(exeData,exeSize,Buffer);
    if (S_OK!=hr) return hr;

    return S_OK;
}

#ifdef _WIN64_
static inline bool IsWow64() { return false; }
#else
static bool IsWow64()
{
    BOOL bIsWow64 = FALSE;

    typedef BOOL(WINAPI *LPFN_ISWOW64PROCESS) (HANDLE,PBOOL);

    LPFN_ISWOW64PROCESS fnIsWow64Process = (LPFN_ISWOW64PROCESS)GetProcAddress(
        GetModuleHandleW(L"kernel32"),"IsWow64Process");

    if (NULL != fnIsWow64Process)
    {
        (*fnIsWow64Process)(GetCurrentProcess(),&bIsWow64);
    }
    return (FALSE!=bIsWow64);
}
#endif

HRESULT WINAPI WinCdArbInstall(PCWSTR ExePath)
{
#ifndef WINCDARB_ALLOW_WOW64_SERVICE
    if (IsWow64())
    {
        // We should not install 32-bit service on 64-bit OS
        return E_FAIL;
    }
#endif

    return ManageSCM(
        SC_MANAGER_CREATE_SERVICE|SC_MANAGER_ENUMERATE_SERVICE,
        (ULONG_PTR)(PVOID)ExePath,
        [](SC_HANDLE scm,ULONG_PTR ulpExePath)
    {
        HRESULT     hr;
        WCHAR       strbuf[MAX_PATH+64];
        PCWSTR      ExePath = (PCWSTR)(PVOID)ulpExePath;

        hr = StopAndDeleteCdArbService(scm);
        if (HRESULT_FROM_WIN32(ERROR_SERVICE_DOES_NOT_EXIST)==hr)
        {
            hr=S_OK;
        }
        if (S_OK!=hr) return hr;

        switch (ulpExePath)
        {
        case InstallExePathCopyThis:
        case InstallExePathCopyTiny:
            hr = CopyCdArbExecutable(strbuf,ARRAY_COUNT(strbuf),(InstallExePathCopyTiny==ulpExePath));
            if (S_OK!=hr) return hr;

            hr = GetCdArbServiceExePath(strbuf,ARRAY_COUNT(strbuf),false);
            if (S_OK!=hr) return hr;

            hr = InstallCdArbService(scm,strbuf);
            break;
        case InstallExePathUseThis:
            hr = GetThisExecutablePath(strbuf,ARRAY_COUNT(strbuf));
            if (S_OK!=hr) return hr;

            hr = InstallCdArbService(scm,strbuf);
            break;
        default:
            hr = InstallCdArbService(scm,ExePath);
            break;
        }


        if (S_OK!=hr) return hr;

        hr = StartCdArbService(scm);
        if (S_OK!=hr) return hr;

        return S_OK;
    });
}

HRESULT WINAPI WinCdArbServiceStart()
{
    return ManageSCM(
        SC_MANAGER_ENUMERATE_SERVICE,0,
        [](SC_HANDLE scm,ULONG_PTR)
    {
        return StartCdArbService(scm);
    });
}

HRESULT WINAPI WinCdArbServiceStop()
{
    return ManageSCM(
        SC_MANAGER_ENUMERATE_SERVICE,0,
        [](SC_HANDLE scm,ULONG_PTR)
    {
        return StopCdArbService(scm);
    });
}

