/*
    WinCdArb - Windows CdRom Device Arbiter service

    Copyright (C) 2007-2019 GuinpinSoft inc <wincdarb@makemkv.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/
#include "winntapi.h"
#include "cdarbsrv.h"

#include <ver_num.h>
#ifdef HAVE_ARCHDEFS_H
#include <archdefs.h>
#endif
#ifdef HAVE_BUILDINFO_H
#include <gen_buildinfo.h>
#endif

#if defined(ARCH_NAME)
#define WINCDARB_ARCH_STRING ARCH_NAME
#elif defined(BUILDINFO_ARCH_NAME)
#define WINCDARB_ARCH_STRING BUILDINFO_ARCH_NAME
#else
#define WINCDARB_ARCH_STRING "(custom build)"
#endif

#if defined(WINCDARB_BUILD_SERVICE_ONLY)
#define WINCDARB_BUILD_STRING "_tiny "
#elif defined(WINCDARB_BUILD_INSTALLER_ONLY)
#define WINCDARB_BUILD_STRING "_inst "
#else
#define WINCDARB_BUILD_STRING " "
#endif

#ifndef WINCDARB_VERSION_STRING
#define WINCDARB_VERSION_STRING WINCDARB_VERSION_NUMBER WINCDARB_BUILD_STRING WINCDARB_ARCH_STRING
#endif

const PCSTR WinCdArbServiceLicense =
    "  License: GNU GPL, source code available at\r\n"
    "  http://www.makemkv.com/wincdarb\r\n"
    "\r\n";

C_ASSERT(sizeof(WINCDARB_CMD_COMMON)<WinIpcMaxMessageLength);
C_ASSERT(sizeof(WINCDARB_CMD_GET_VERSION_RESPONSE)<WinIpcMaxMessageLength);
C_ASSERT(sizeof(WINCDARB_CMD_GET_HANDLE_REQUEST)<WinIpcMaxMessageLength);
C_ASSERT(sizeof(WINCDARB_CMD_GET_HANDLE_RESPONSE)<WinIpcMaxMessageLength);

// This is a first check, it prevents a rougue client from making us open an arbitrary
// path. This check is NOT sufficient by itself, as a process with medium privileges can
// create a symlink with permitted name pointing to any object in a system. We perform the
// real check for a SCSI MMC device later.
static bool IsAllowedDeviceName(ULONG NameLength,PCWSTR Name)
{
    // \Device\CdRomXXXXXXXXXXX
    if (NameLength>13)
    {
        if (0==memcmp(Name,L"\\Device\\CdRom",13*sizeof(WCHAR)))
        {
            for (unsigned int i=13; i<NameLength; i++)
            {
                if ((Name[i]<'0') || (Name[i]>'9'))
                {
                    return false;
                }
            }
            return true;
        }
    }

    // nothing else is allowed
    return false;
}

// A check for inquiry data. By the time this function it is called, we have
// verified that the opened handle is a SCSI device handle, and we got a valid
// inquiry data. This function ensures that the inquiry data comes from a
// MMC device.
static bool IsAllowedDeviceInquiryData(PUCHAR Data,ULONG DataLength)
{
    if (DataLength<0x24) return false;

    if (Data[0]!=0x05) return false;
    if ((Data[1]&0x80)==0) return false;

    return true;
}

static inline NTSTATUS StatusFromError(DWORD Error)
{
    if (Error==0) {
        return STATUS_SINGLE_STEP;
    } else {
        return HRESULT_FROM_WIN32(Error);
    }
}

static NTSTATUS ScsiDeviceInquiry(HANDLE hDevice,PUCHAR Buffer,PULONG BufferSize)
{
    NTSTATUS        status;
    PVOID           page;
    DWORD           size;
    ULONG           length;
    SCSI_ADDRESS    scsiAddress;
    struct _spd
    {
        SCSI_PASS_THROUGH_DIRECT    sp;
        UCHAR                       sensebuf[64];
        UCHAR                       this_field_IS_required[8];
    } spd;

    C_ASSERT(FIELD_OFFSET(struct _spd,sp) == 0);

    length = *BufferSize;

    page = VirtualAlloc(NULL,length,MEM_COMMIT,PAGE_READWRITE);
    if (NULL==page)
    {
        return STATUS_NO_MEMORY;
    }

    if (FALSE==DeviceIoControl(hDevice,IOCTL_SCSI_GET_ADDRESS,
        NULL,0,
        &scsiAddress,sizeof(scsiAddress),
        &size,NULL))
    {
        memset(&scsiAddress,0,sizeof(scsiAddress));
    }

    memset(&spd,0,sizeof(spd));
    spd.sp.Length = sizeof(SCSI_PASS_THROUGH_DIRECT);
    spd.sp.PathId = scsiAddress.PathId;
    spd.sp.TargetId = scsiAddress.TargetId;
    spd.sp.Lun = scsiAddress.Lun;
    spd.sp.SenseInfoLength = sizeof(spd.sensebuf);
    spd.sp.DataIn = SCSI_IOCTL_DATA_IN;
    spd.sp.DataTransferLength = length;
    spd.sp.TimeOutValue = 30;
    spd.sp.DataBuffer = page;
    spd.sp.SenseInfoOffset = FIELD_OFFSET(struct _spd,sensebuf);

    spd.sp.CdbLength = 6;
    spd.sp.Cdb[0] = 0x12;
    spd.sp.Cdb[3] = (UCHAR)(length>>8);
    spd.sp.Cdb[4] = (UCHAR)(length);

    if (FALSE==DeviceIoControl(hDevice,IOCTL_SCSI_PASS_THROUGH_DIRECT,
        &spd,sizeof(spd),
        &spd,sizeof(spd),
        &size,NULL))
    {
        status = StatusFromError(GetLastError());
    } else {
        status = STATUS_SUCCESS;
    }

    if ((status==STATUS_SUCCESS) &&
        (size>=sizeof(SCSI_PASS_THROUGH_DIRECT)) &&
        (0==spd.sp.ScsiStatus))
    {
        length = spd.sp.DataTransferLength;
        if (length>(*BufferSize))
        {
            length = (*BufferSize);
        }
        memcpy(Buffer,page,length);
    } else {
        status = STATUS_INVALID_HANDLE;
    }

    VirtualFree(page,*BufferSize,MEM_FREE);

    *BufferSize = length;

    return status;
}

// Opens a (presumably) CdRom device and checks that the opened
// handle is in fact a handle for SCSI MMC device. The returned
// handle has access rigts sufficient to read data and issue
// SPTI IOCTLs.
NTSTATUS OpenCdRomDevice(PHANDLE Handle,PCWSTR DeviceName)
{
    NTSTATUS            status;
    OBJECT_ATTRIBUTES   oa;
    IO_STATUS_BLOCK     ioStatusBlock;
    UNICODE_STRING      ustrName;
    HANDLE              hDevice;
    UCHAR               buffer[0x24];
    DWORD               size;

    memset(&ioStatusBlock,0,sizeof(ioStatusBlock));

    RtlInitUnicodeString(&ustrName,(PWSTR)DeviceName);
    InitializeObjectAttributes(&oa,&ustrName,0,NULL,NULL);

    status = NtOpenFile(
        &hDevice,
        SYNCHRONIZE | FILE_READ_DATA | FILE_WRITE_DATA | STANDARD_RIGHTS_READ,
        &oa,
        &ioStatusBlock,
        FILE_SHARE_READ | FILE_SHARE_WRITE,
        FILE_NO_INTERMEDIATE_BUFFERING |
        FILE_NON_DIRECTORY_FILE |
        FILE_FLAG_OPEN_REPARSE_POINT );

    if (STATUS_SUCCESS!=status) return status;

    size = sizeof(buffer);
    status = ScsiDeviceInquiry(hDevice,buffer,&size);
    if (STATUS_SUCCESS!=status)
    {
        // Under attack: This is not a SCSI device!
        NtClose(hDevice);
        return status;
    }

    if (false==IsAllowedDeviceInquiryData(buffer,size))
    {
        // Under attack: This is a handle for a non-MMC device!
        status = STATUS_ACCESS_VIOLATION;
        NtClose(hDevice);
        return status;
    }

    *Handle = hDevice;
    return STATUS_SUCCESS;
}

static NTSTATUS DoCmdGetHandle(PHANDLE Handle,ULONG ProcessId,ULONG NameLength,PWSTR Name)
{
    HANDLE      hProcess,hDevice;
    NTSTATUS    status;

    if (NameLength>=WINCDARB_MAX_DEVICE_NAME_LENGTH) return STATUS_INVALID_PARAMETER;
    Name[NameLength] = 0;

    if (true!=IsAllowedDeviceName(NameLength,Name))
    {
        return STATUS_ACCESS_VIOLATION;
    }

    hProcess = OpenProcess(PROCESS_DUP_HANDLE,FALSE,ProcessId);
    if (NULL==hProcess)
    {
        return StatusFromError(GetLastError());
    }

    status = OpenCdRomDevice(&hDevice,Name);
    if (STATUS_SUCCESS!=status)
    {
        NtClose(hProcess);
        return status;
    }

    if (FALSE==DuplicateHandle(
        GetCurrentProcess(),hDevice,
        hProcess,Handle,
        0,FALSE,
        DUPLICATE_CLOSE_SOURCE|DUPLICATE_SAME_ACCESS))
    {
        status = StatusFromError(GetLastError());
    } else {
        status = STATUS_SUCCESS;
    }

    NtClose(hProcess);
    return status;
}

static BOOLEAN NTAPI WinCdArbProcessMessage(PVOID /*Context*/,ULONG ProcessId,PUCHAR Message,PULONG MessageSize)
{
    ULONG size;

    size = *MessageSize;
    if (size < sizeof(WINCDARB_CMD_COMMON)) return FALSE;

    switch (((WINCDARB_CMD_COMMON*)Message)->CommandCode)
    {
    case WinCdArbCmdGetVersion:
        {
            WINCDARB_CMD_GET_VERSION_RESPONSE* p_res = (WINCDARB_CMD_GET_VERSION_RESPONSE*)Message;
            p_res->Version = 1;

            C_ASSERT(sizeof(WINCDARB_VERSION_STRING)<sizeof(p_res->VersionString));
            memcpy(p_res->VersionString,WINCDARB_VERSION_STRING,sizeof(WINCDARB_VERSION_STRING));

            size = sizeof(WINCDARB_CMD_GET_VERSION_RESPONSE);
        }
        break;
    case WinCdArbCmdGetHandle:
        {
            NTSTATUS    status;
            HANDLE      handle = NULL;
            WINCDARB_CMD_GET_HANDLE_REQUEST*  p_req = (WINCDARB_CMD_GET_HANDLE_REQUEST*)Message;
            WINCDARB_CMD_GET_HANDLE_RESPONSE* p_res = (WINCDARB_CMD_GET_HANDLE_RESPONSE*)Message;

            status = DoCmdGetHandle(&handle,ProcessId,p_req->DeviceNameLength,p_req->DeviceName);

            p_res->ErrorCode = status;
            p_res->Handle = (ULONG_PTR)(PVOID)handle;
            size = sizeof(WINCDARB_CMD_GET_HANDLE_RESPONSE);
        }
        break;
    default:
        size = 0;
        break;
    }
    if ((0==size) || (size>WinIpcMaxMessageLength))
    {
        return FALSE;
    }

    *MessageSize = size;
    return TRUE;
}

void WINAPI WinCdArbServerInit(void)
{
    WinIpcClearExitServerLoopFlag();
}

HRESULT NTAPI WinCdArbServerLoop(void)
{
    HANDLE      hPort;
    NTSTATUS    status;

    status = WinIpcCreatePort(&hPort,WINCDARB_PORT_NAME);
    if (STATUS_SUCCESS!=status) return status;

    status = WinIpcServerLoop(hPort,WinCdArbProcessMessage,NULL);

    NtClose(hPort);

    return status;
}

const char* WINAPI WinCdArbVersionString()
{
    return WINCDARB_VERSION_STRING;
}

