/*
    WinCdArb - Windows CdRom Device Arbiter service

    Copyright (C) 2007-2019 GuinpinSoft inc <wincdarb@makemkv.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/
#include "winntapi.h"
#include "ntipc.h"
#include <wincdarb/winipc.h>

static LONG volatile g_terminate = 0;

static void InterlockedSet(LONG volatile * _Value,LONG NewValue)
{
    LONG old;
    do
    {
        old = *_Value;
    } while (old!=InterlockedCompareExchange(_Value,NewValue,old));
}

void WINAPI WinIpcClearExitServerLoopFlag()
{
    InterlockedSet(&g_terminate,0);
}

void WINAPI WinIpcSetExitServerLoopFlag()
{
    InterlockedIncrement(&g_terminate);
}

NTSTATUS WINAPI WinIpcCreatePort(PHANDLE PortHandle,PCWSTR PortName)
{
    UNICODE_STRING ustrName;
    OBJECT_ATTRIBUTES oa;

    RtlInitUnicodeString(&ustrName,(PWSTR)PortName);
    InitializeObjectAttributes(&oa,&ustrName,0,NULL,NULL);

    return NtCreatePort(PortHandle,&oa,IPC_MAX_DATA_LENGTH,IPC_MAX_DATA_LENGTH,NULL);
}

NTSTATUS WINAPI WinIpcServerLoop(HANDLE PortHandle,WinIpcProcessMessageRoutine ProcessMessage,PVOID Context)
{
    NTSTATUS        status;
    BOOLEAN         r;
    ULONG           headerLength;
    ULONG           msgSize;
    HANDLE          hConnectHandle;
    LARGE_INTEGER   timeout;
    PUCHAR          msgBuf;
    struct _msg
    {
        IPC_MESSAGE hdr;
        UCHAR       buffer[(IPC_MAX_HEADER_LENGTH+IPC_MAX_DATA_LENGTH)-sizeof(IPC_MESSAGE)];
        UCHAR       justincasenotneededreally[4];
    } msg;

    while (InterlockedIncrement(&g_terminate)<=1)
    {
        InterlockedDecrement(&g_terminate);

        timeout.QuadPart = ((ULONGLONG)0) - 5000*10000ll; // 5000 ms

        status = NtReplyWaitReceivePortEx(
            PortHandle,
            NULL,
            NULL,
            &msg.hdr,
            &timeout);

        if (STATUS_TIMEOUT==status) continue;
        if (STATUS_SUCCESS!=status) return status;

        if (msg.hdr.MessageType!=IPC_MESSAGE_TYPE_CONNECTION_REQUEST)
        {
            continue;
        }

        // we can get 64-bit msg in wow32 process
        headerLength = msg.hdr.Length - msg.hdr.DataLength;
        if ((headerLength!=IPC_HEADER_LENGTH_32) &&
            (headerLength!=IPC_HEADER_LENGTH_64))
        {
            return STATUS_INVALID_PARAMETER;
        }

        // these could not happen under any circumstances
        if (msg.hdr.DataLength>IPC_MAX_DATA_LENGTH) return STATUS_STACK_OVERFLOW;
        if (msg.hdr.Length>(headerLength+IPC_MAX_DATA_LENGTH)) return STATUS_STACK_OVERFLOW;
        if (IPC_HEADER_LENGTH_64==headerLength)
        {
            if (msg.hdr.UniqueProcessIdHigh64!=0) return STATUS_INVALID_HANDLE;
        }

        msgBuf = ((PUCHAR)&msg.hdr)+headerLength;
        msgSize = msg.hdr.DataLength;

        // Avoid data leak. This IS important.
        memset(msgBuf+msgSize,0,IPC_MAX_DATA_LENGTH-msgSize);

        r = (*ProcessMessage)(
            Context,
            msg.hdr.UniqueProcessId,
            msgBuf,
            &msgSize);

        if (FALSE==r)
        {
            msgSize = 0;
        }
        msg.hdr.Length = (USHORT)(headerLength + msgSize);
        msg.hdr.DataLength = (USHORT)msgSize;

        status = NtAcceptConnectPort(&hConnectHandle,NULL,&msg.hdr,r,NULL,NULL);

        if ((STATUS_SUCCESS==status) && (FALSE!=r))
        {
            // Existense of NtReplyWaitReceivePortEx implies that NtCompleteConnectPort is a noop.
            // Keep it just in case, to be on a safe side...
            NtCompleteConnectPort(hConnectHandle);
            NtClose(hConnectHandle);
        }
    }
    return STATUS_SUCCESS;
}

